==> fispro_3.4.orig.tar.gz: fispro-3.4/cpp/hfp/hfpintri.cpp <==
//***********************************************************************
//
//
//                              HFPINTRI.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : INHFP class functions depending on  MF shape 
//        writen for the triangular one.
//        Used by FISPRO, part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fispro
//
//**********************************************************************

#include "hfpin.h"


void  INHFP::InitMfBordTri(double * v)
  //***********************************
{
  double left, right;
  int i;

  Fp = new MF *[Nmf];

  for(i = 0; i < Nmf; i++)
    {
      if(i == 0) left = v[0]*2 - v[1];           // out of range
      else left = v[i-1];
      if(i == Nmf - 1) right = v[i]*2 - v[i-1];    // out of range
      else right = v[i+1];
      Fp[i] = new MFTRI(left, v[i], right);
    }
}
void  INHFP::InitMfBordTrap(double * v)
  //***********************************
{
  double left, right;
  int i;

  Fp = new MF *[Nmf];

  for(i = 0; i < Nmf; i++)
    {
      if(i == 0) left = -INFINI_HFP;           // out of range
      else left = v[i-1];
      if(i == Nmf - 1) right = INFINI_HFP;
      else right = v[i+1];
      Fp[i] = new MFTRI(left, v[i], right);
    }
}

void INHFP::MfInstall(int n, int Strap)
  //***********************************
{
  int i;
  double left, centre, right, diff;

  if(active == false) return;

  // correcting memory leaks - delete Fp before changing Nmf
  if (Fp !=NULL)
    for (i=0;i<Nmf;i++)
      delete Fp[i];
    
  Nmf = n;
  diff = ValSup - ValInf;
  // if(diff < EPSILON) diff = 1.0;

  if(Nmf == 1)
    {
      //   delete Fp[0];
      Fp[0] = new MFUNIV(ValInf, ValSup);
      return;
    }

  for(i = 0; i < Nmf; i++)
    {
      if(i == 0) left = -INFINI_HFP;           // out of range
      else left = Vertices[Nmf - 1][i].l * diff + ValInf;

      centre =  Vertices[Nmf - 1][i + 1].l * diff + ValInf ;

      if(i == Nmf - 1) right = INFINI_HFP;    // out of range
      else right = Vertices[Nmf - 1][i + 2].l  * diff + ValInf;

      // delete Fp[i];

      if(Strap)
	{	
	  if(i == 0)
	    Fp[i] = new MFTRAPINF(ValInf, centre, right);
	  else 
	    if(i == Nmf - 1) Fp[i] = new MFTRAPSUP(left, centre, ValSup);
	    else Fp[i] = new MFTRI(left, centre, right);
	}
      else Fp[i] = new MFTRI(left, centre, right);
    }
}

void INHFP::SelectMf(double * Vs)
  //*****************************
{
  int i, j, k, before, dim, iloop, index;
  double varm, weightm, val, max, min;
  double *var, *weight, *vdensity, *VarMf, *mass;
  double g1, c1, d1, mas1, g2, c2, d2, mas2;
  
  MFTRI ** Fpt; 
  MFTRI ** Original; 

  Fpt = (MFTRI **) Fp;
  for(i = 0; i < Nmf; i++) Fpt[i] = (MFTRI *) Fp[i];

  MFTRI S1(Fpt[0]);        // to save original mf
  MFTRI S2(Fpt[0]);
  MFTRI Sprec(Fpt[0]);
  MFTRI Ssuiv(Fpt[0]);

  NmfInit = Nmf;
  var = weight = vdensity = VarMf = mass = NULL;

  Original  = new MFTRI *[Nmf];
  mass = new double [Nmf];

  for(i = 0; i < Nmf; i++)
    {
      Original[i]  = new MFTRI(Fpt[i]) ;
      mass[i] = Occur[i];	
    }
  
  dim = ((NmfVar > Nvertices + 1) ? NmfVar : (Nvertices + 1));

  var = new double [dim+1];
  weight = new double [dim+1];
  vdensity = new double [dim];
  VarMf = new double [dim];
  
  iloop = 1;
  while(iloop < NmfInit-1)  
    {
      Nmf --;  // For sum of distances computation
      iloop ++;
      index = -1;
      max = -1;
      min = INFINI_HFP;

      // Heterogeneity index is only done for the last steps
      if( Vs && ( Nmf < NmfVar || Nmf < Nvertices + 1) )
	VarianceNP(Vs, 0, var[0], weight[0]);

      for(i = 0; i < Nmf; i++)     // one step
	{
	  S1 = *(Fpt[i]);              // save original mfs
	  S2 = *(Fpt[i+1]);
	  
	  if( Vs && Nmf < NmfVar) 
	    {
	      Nmf++;     // To access the last mf 
	      VarianceNP(Vs, i+1, var[i+1], weight[i+1]);
	      Nmf--;
	    }
	  
	  // Merging
	  Fpt[i]->ValParam(g1, c1, d1);
	  Fpt[i+1]->ValParam(g2, c2, d2);
	  mas1 = mass[i];
	  mas2 = mass[i+1];
	  
	  delete Fpt[i];
	  Fpt[i] = new MFTRI(g1, ((c1*mas1)+(c2*mas2))/(mas1+mas2), d2); 
	  
	  if(i)          // Adjust the mf before the merged one
	    { 
	      Sprec =  *(Fpt[i-1]);
	      
	      Fpt[i-1]->ValParam(g1, c1, d1);
	      Fpt[i]->ValParam(g2, c2, d2);
	      delete Fpt[i-1];
	      Fpt[i-1] = new MFTRI(g1, c1, c2); 
	    }
	  
	  if(i < Nmf - 1) // Adjust the mf following the merged one
	    {
	      Ssuiv = *(Fpt[i+2]); 

	      Fpt[i]->ValParam(g1, c1, d1);
	      Fpt[i+2]->ValParam(g2, c2, d2);
	      delete Fpt[i+2];
	      //to be deleted later
	      Fpt[i+2] = new MFTRI(c1, c2, d2); 
	    }
	  
	  // Update (shift) of the partition
	  for(j = i + 1; j < Nmf; j++) *(Fpt[j]) = *(Fpt[j + 1]);
	  
	  // Sum of distances after the merging
	  before = 0;
	  val = DistSum(i, before);
	  
	  // i is the merged mf
	  if (Vs) VarianceNP(Vs, i, varm, weightm);       
	  
	  if( Vs && Nmf < NmfVar)   
	    { 
	      Density(vdensity);
	      
	      if( ((var[i] < VarThres) && (var[i+1] < VarThres) && (varm > VarThres))
		  || ((var[i] < VarThres) && (var[i+1] > VarThres))
		  || ((var[i] > VarThres) && (var[i+1] < VarThres))  )
		{
		  varm *= Penalty;
		  val /= Penalty;
		}
	    }
	  
	  if(VarCrit)   // Minimum of variance
	    {
	      if(varm < min)
		{
		  min = varm;
		  index = i;
		}
	    }
	  else         // Maximum of sum of distances
	    {  
	      if(val > max) 
		{
		  max = val;
		  index = i;
		}
	    }      

	  // Restoration
	  for(j = Nmf; j > i + 1 ; j--)   *(Fpt[j]) = *(Fpt[j - 1]); 
	  *(Fpt[i]) = S1; 
	  *(Fpt[i+1]) = S2;
	  if(i)  *(Fpt[i-1]) = Sprec; 
	  if(i < Nmf - 1)  *(Fpt[i+2]) = Ssuiv; 
	  
	}	
    
      // Merge the best configuration
      Fpt[index]->ValParam(g1, c1, d1);
      Fpt[index+1]->ValParam(g2, c2, d2);
      mas1 = mass[index];
      mas2 = mass[index+1];

      delete Fpt[index];
      //to be deleted later
      Fpt[index] = new MFTRI(g1, ((c1*mas1)+(c2*mas2))/(mas1+mas2), d2); 
      mass[index] = EvalMass(index);
      
      for(j = index + 1; j < Nmf; j++) 
	{
	  *(Fpt[j]) = *(Fpt[j + 1]);
	  mass[j] = mass[j+1];
	}
      
      if(index)   
	{    
	  Fpt[index-1]->ValParam(g1, c1, d1);
	  Fpt[index]->ValParam(g2, c2, d2);
	  delete Fpt[index-1];
	  Fpt[index-1] = new MFTRI(g1, c1, c2); //to be deleted later
	  mass[index - 1] = EvalMass(index - 1);
	}
      
      if(index < Nmf - 1) 
	{
	  Fpt[index]->ValParam(g1, c1, d1);
	  Fpt[index+1]->ValParam(g2, c2, d2);
	  delete Fpt[index+1];//to be deleted later
	  Fpt[index+1] = new MFTRI(c1, c2, d2); 
	  mass[index + 1] = EvalMass(index + 1);
	}
      
      // Store configuration (vertices).
      if (Nmf < Nvertices + 1) 
	{
	  double med, mean, Tmin, Tmax;
	  Density(vdensity);
	  StatArray(vdensity, Nmf, 0, med, mean, val, Tmin, Tmax, 0);
	  if (Vs) varm = Variance(Vs, VarMf);
	  else varm=0.0;
	  Vertices[Nmf-1][0].l = val;
	  Vertices[Nmf-1][0].r = varm;
	  for(k = 0; k < Nmf; k++) 
	    {
	      Fpt[k]->Kernel(Vertices[Nmf-1][k+1].l, Vertices[Nmf-1][k+1].r);
	      Vertices[Nmf-1][k+1].p = 0;
	      if(varm && VarMf[k] < VarThres) Vertices[Nmf-1][k+1].p = 1;
	    }
	}
    } // end while()  

  delete [] var;  delete [] VarMf;  delete [] weight; delete [] vdensity; delete [] mass;
  
  // memory cleaning - bch 17 July 2009
  for(i = 0; i < NmfInit; i++) delete Fpt[i];
  for(i = 0; i < NmfInit; i++)  Fpt[i] = Original[i];  // Restoration
  Nmf = NmfInit;
  
  delete [] Original;
} // End of INHFP::SelectMf()

//***************************   End of HFPINTRI.CPP   *********************

==> fispro_3.6.orig.tar.gz: fispro-3.6/cpp/hfp/hfpintri.cpp <==
//***********************************************************************
//
//
//                              HFPINTRI.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : INHFP class functions depending on  MF shape 
//        writen for the triangular one.
//        Used by FISPRO, part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fispro
//
//**********************************************************************

#include "hfpin.h"


void  INHFP::InitMfBordTri(double * v)
  //***********************************
{
  double left, right;
  int i;

  Fp = new MF *[Nmf];

  for(i = 0; i < Nmf; i++)
    {
      if(i == 0) left = -0.05;
      else left = v[i-1];
      if(i == Nmf - 1) right = 1.05;    // out of range
      else right = v[i+1];
      Fp[i] = new MFTRI(left, v[i], right);
    }
}
void  INHFP::InitMfBordTrap(double * v)
  //***********************************
{
  double left, right;
  int i;

  Fp = new MF *[Nmf];

  for(i = 0; i < Nmf; i++)
    {
      if(i == 0) left = -INFINI_HFP;           // out of range
      else left = v[i-1];
      if(i == Nmf - 1) right = INFINI_HFP;
      else right = v[i+1];
      Fp[i] = new MFTRI(left, v[i], right);
    }
}

void INHFP::MfInstall(int n, int Strap)
  //***********************************
{
  int i;
  double left, centre, right, diff;
  char * name=NULL;

  if(active == false) return;

  if (Fp !=NULL)
    for (i=0;i<Nmf;i++) delete Fp[i];
    
  Nmf = n;
  diff = ValSup - ValInf;

  if(Nmf == 1)
    {
      Fp[0] = new MFUNIV(ValInf, ValSup);
      return;
    }

  for(i = 0; i < Nmf; i++)
    {
      if(i == 0) left = -INFINI_HFP;           // out of range
      else left = Vertices[Nmf - 1][i].l * diff + ValInf;

      centre =  Vertices[Nmf - 1][i + 1].l * diff + ValInf ;

      if(i == Nmf - 1) right = INFINI_HFP;    // out of range
      else right = Vertices[Nmf - 1][i + 2].l  * diff + ValInf;

      if(Strap)
	{	
	  if(i == 0)
	    Fp[i] = new MFTRAPINF(ValInf, centre, right);
	  else 
	    if(i == Nmf - 1) Fp[i] = new MFTRAPSUP(left, centre, ValSup);
	    else Fp[i] = new MFTRI(left, centre, right);
	}
      else Fp[i] = new MFTRI(left, centre, right);
      name=new char[6];
      sprintf(name,"MF%d",i+1);
      Fp[i]->SetName(name);
      delete [] name;
    }
}

void INHFP::SelectMf(void)
  //**********************
{
  int i, j, k, before, dim, iloop, index;
  double val, max, min, partc, parte;
  double g1, c1, d1, mas1, g2, c2, d2, mas2, *mass;
  
  MFTRI ** Fpt; 
  MFTRI ** Original; 

  Fpt = (MFTRI **) Fp;
  for(i = 0; i < Nmf; i++) Fpt[i] = (MFTRI *) Fp[i];

  MFTRI S1(Fpt[0]);        // to save original mf
  MFTRI S2(Fpt[0]);
  MFTRI Sprec(Fpt[0]);
  MFTRI Ssuiv(Fpt[0]);

  NmfInit = Nmf;
  mass = NULL;

  Original  = new MFTRI *[Nmf];
  mass = new double [Nmf];

  for(i = 0; i < Nmf; i++)
    {
      Original[i]  = new MFTRI(Fpt[i]) ;
      mass[i] = Occur[i];
   }

  dim = Nvertices + 1;

  partc = parte = 0.;

  iloop = 1;
  while(iloop < NmfInit-1)  
    {
      Nmf --;  // For sum of distances computation
      iloop ++;
      index = -1;
      max = -1;
      min = INFINI_HFP;

      for(i = 0; i < Nmf; i++)     // one step
	{
	  S1 = *(Fpt[i]);              // save original mfs
	  S2 = *(Fpt[i+1]);
	  
	  // Merging
	  Fpt[i]->ValParam(g1, c1, d1);
	  Fpt[i+1]->ValParam(g2, c2, d2);
	  mas1 = mass[i];
	  mas2 = mass[i+1];
	  
	  delete Fpt[i];
	  Fpt[i] = new MFTRI(g1, ((c1*mas1)+(c2*mas2))/(mas1+mas2), d2); 
	  
	  if(i)          // Adjust the mf before the merged one
	    { 
	      Sprec =  *(Fpt[i-1]);
	      
	      Fpt[i-1]->ValParam(g1, c1, d1);
	      Fpt[i]->ValParam(g2, c2, d2);
	      delete Fpt[i-1];
	      Fpt[i-1] = new MFTRI(g1, c1, c2); 
	    }
	  
	  if(i < Nmf - 1) // Adjust the mf following the merged one
	    {
	      Ssuiv = *(Fpt[i+2]); 

	      Fpt[i]->ValParam(g1, c1, d1);
	      Fpt[i+2]->ValParam(g2, c2, d2);
	      delete Fpt[i+2];
	      Fpt[i+2] = new MFTRI(c1, c2, d2); 
	    }
	  
	  // Update (shift) of the partition
	  for(j = i + 1; j < Nmf; j++) *(Fpt[j]) = *(Fpt[j + 1]);
	  
	  // Sum of distances after the merging
	  before = 0;
	  val = DistSum(i, before);
	  
	  if(val > max) 
	    {
	      max = val;
	      index = i;
	    }

	  // Restoration
	  for(j = Nmf; j > i + 1 ; j--)   *(Fpt[j]) = *(Fpt[j - 1]); 
	  *(Fpt[i]) = S1; 
	  *(Fpt[i+1]) = S2;
	  if(i)  *(Fpt[i-1]) = Sprec; 
	  if(i < Nmf - 1)  *(Fpt[i+2]) = Ssuiv; 
	}	
    
      // Merge the best configuration
      Fpt[index]->ValParam(g1, c1, d1);
      Fpt[index+1]->ValParam(g2, c2, d2);
      mas1 = mass[index];
      mas2 = mass[index+1];

      delete Fpt[index];
      //to be deleted later
      Fpt[index] = new MFTRI(g1, ((c1*mas1)+(c2*mas2))/(mas1+mas2), d2); 
      mass[index] = EvalMass(index);
      
      for(j = index + 1; j < Nmf; j++) 
	{
	  *(Fpt[j]) = *(Fpt[j + 1]);
	  mass[j] = mass[j+1];
	}
      
      if(index)   
	{    
	  Fpt[index-1]->ValParam(g1, c1, d1);
	  Fpt[index]->ValParam(g2, c2, d2);
	  delete Fpt[index-1];
	  Fpt[index-1] = new MFTRI(g1, c1, c2); //to be deleted later
	  mass[index - 1] = EvalMass(index - 1);
	}
      
      if(index < Nmf - 1) 
	{
	  Fpt[index]->ValParam(g1, c1, d1);
	  Fpt[index+1]->ValParam(g2, c2, d2);
	  delete Fpt[index+1];//to be deleted later
	  Fpt[index+1] = new MFTRI(c1, c2, d2); 
	  mass[index + 1] = EvalMass(index + 1);
	}
      
      // Store configuration (vertices).
      if (Nmf < Nvertices + 1) 
	{
	  double *tab;
	  tab = new double[Nmf];
	  GetMfCenters(tab);
	  ComputeIndices(tab, Nmf, partc, parte);
	  delete [] tab;

	  Vertices[Nmf-1][0].l = partc;
	  Vertices[Nmf-1][0].r = parte;
	  for(k = 0; k < Nmf; k++) 
	    Fpt[k]->Kernel(Vertices[Nmf-1][k+1].l, Vertices[Nmf-1][k+1].r);
	}
    } // end while()  

  delete [] mass;
  
  for(i = 0; i < NmfInit; i++) delete Fpt[i];
  for(i = 0; i < NmfInit; i++)  Fpt[i] = Original[i];  // Restoration
  Nmf = NmfInit;
  
  delete [] Original;
} // End of INHFP::SelectMf()

//***************************   End of HFPINTRI.CPP   *********************
